@@ -4,14 +4,11 @@ from inspect import isfunction
 from functools import partial
 import torch.nn.functional as F
 
+from math import sqrt
 from einops import rearrange, repeat
 
-# constants
-
-MAX_NUM_MSA = 20
-NUM_AMINO_ACIDS = 21
-NUM_EMBEDDS_TR  = 1280 # best esm model 
-DISTOGRAM_BUCKETS = 37
+import alphafold2_pytorch.constants as constants
+from alphafold2_pytorch.reversible import ReversibleSequence, SequentialSequence
 
 # helpers
 
@@ -31,9 +28,9 @@ class PreNorm(nn.Module):
         self.norm = nn.LayerNorm(dim)
         self.fn = fn
 
-    def forward(self, x, **kwargs):
+    def forward(self, x, *args, **kwargs):
         x = self.norm(x)
-        return self.fn(x, **kwargs)
+        return self.fn(x, *args, **kwargs)
 
 class GEGLU(nn.Module):
     def forward(self, x):
@@ -111,7 +108,12 @@ class AxialAttention(nn.Module):
         self.attn_height = Attention(**kwargs)
 
     def forward(self, x, context = None, mask = None, context_mask = None):
-        b, h, w, d = x.shape
+        b, n, d = x.shape
+        w = h = int(sqrt(n))
+
+        x = x.reshape(b, h, w, d)
+        mask = mask.reshape(b, h, w)
+
         w_mask = h_mask = w_context = h_context = w_context_mask = h_context_mask = None
 
         if exists(mask):
@@ -132,7 +134,8 @@ class AxialAttention(nn.Module):
         h_out = self.attn_height(h_x, mask = h_mask, context = h_context, context_mask = h_context_mask)
         h_out = rearrange(h_out, '(b h) w d -> b h w d', h = h, w = w)
 
-        return w_out + h_out
+        out = w_out + h_out
+        return rearrange(out, 'b h w d -> b (h w) d')
 
 # main class
 
@@ -146,61 +149,79 @@ class Alphafold2(nn.Module):
         heads = 8,
         dim_head = 64,
         pos_token = 3,
-        num_tokens = NUM_AMINO_ACIDS,
-        num_embedds = NUM_EMBEDDS_TR,
+        num_tokens = constants.NUM_AMINO_ACIDS,
+        num_embedds = constants.NUM_EMBEDDS_TR,
         attn_dropout = 0.,
-        ff_dropout = 0.
+        ff_dropout = 0.,
+        reversible = False
     ):
         super().__init__()
         self.token_emb = nn.Embedding(num_tokens, dim)
-
         self.pos_emb = nn.Embedding(max_seq_len, dim)
 
         # multiple sequence alignment position embedding
 
         self.msa_pos_emb = nn.Embedding(max_seq_len, dim)
-        self.msa_num_pos_emb = nn.Embedding(MAX_NUM_MSA, dim)
+        self.msa_num_pos_emb = nn.Embedding(constants.MAX_NUM_MSA, dim)
 
         # custom embedding projection
 
         self.embedd_project = nn.Linear(num_embedds, dim)
 
+        # main trunk modules
+
         wrapper = partial(PreNorm, dim)
 
-        self.layers = nn.ModuleList([])
+        layers = nn.ModuleList([])
         for _ in range(depth):
-            self.layers.append(nn.ModuleList([
+            layers.append(nn.ModuleList([
                 wrapper(AxialAttention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
+                wrapper(FeedForward(dim = dim, dropout = ff_dropout)),
                 wrapper(Attention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
                 wrapper(FeedForward(dim = dim, dropout = ff_dropout)),
             ]))
 
-
-        self.msa_layers = nn.ModuleList([])
-        for _ in range(depth):
-            self.msa_layers.append(nn.ModuleList([
+            layers.append(nn.ModuleList([
                 wrapper(Attention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
+                wrapper(FeedForward(dim = dim, dropout = ff_dropout)),
                 wrapper(Attention(dim = dim, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
-                wrapper(FeedForward(dim = dim, dropout = ff_dropout))
+                wrapper(FeedForward(dim = dim, dropout = ff_dropout)),
             ]))
 
-        self.norm = nn.LayerNorm(dim)
-        self.to_distogram_logits = nn.Linear(dim, DISTOGRAM_BUCKETS)
+        if not reversible:
+            layers = nn.ModuleList(list(map(lambda t: t[:3], layers))) # remove last feed forward if not reversible
+
+        trunk_class = SequentialSequence if not reversible else ReversibleSequence
+        self.net = trunk_class(layers)
+
+        # to output
+
+        self.to_distogram_logits = nn.Sequential(
+            nn.LayerNorm(dim),
+            nn.Linear(dim, constants.DISTOGRAM_BUCKETS)
+        )
 
     def forward(self, seq, msa = None, embedds = None, mask = None, msa_mask = None):
         n, device = seq.shape[1], seq.device
 
-        # unpack (AA_cod, atom_pos)
-        if isinstance(seq, list):
+        # unpack (AA_code, atom_pos)
+
+        if isinstance(seq, (list, tuple)):
             seq, seq_pos = seq
+
         # embed main sequence
+
         x = self.token_emb(seq)
         x += self.pos_emb(torch.arange(n, device = device))[None, ...]
         x = x[:, :, None, :] + x[:, None, :, :] # create pair-wise residue embeds
         x_mask = mask[:, :, None] * mask[:, None, :]
 
+        x = rearrange(x, 'b i j d -> b (i j) d')
+        x_mask = rearrange(x_mask, 'b i j -> b (i j)')
+
         # embed multiple sequence alignment
 
+        m = None
         if exists(msa):
             msa_shape = msa.shape
 
@@ -212,7 +233,6 @@ class Alphafold2(nn.Module):
 
         elif exists(embedds):
             m = self.embedd_project(embedds)
-            # pairwise mat - maybe repeating stuff? 
             m = m[:, :, None, :] + m[:, None, :, :]
             m = rearrange(m, 'b m n d -> b (m n) d')
 
@@ -221,44 +241,8 @@ class Alphafold2(nn.Module):
 
         # trunk
 
-        for ((attn, cross_attn, ff), (msa_attn, msa_cross_attn, msa_ff)) in zip(self.layers, self.msa_layers):
-
-            # self attention
-
-            x = attn(x, mask = x_mask) + x
-
-            if exists(msa):
-                m = msa_attn(m, mask = msa_mask) + m
-
-                # cross attention
-
-                x = rearrange(x, 'b i j d -> b (i j) d')
-                x_mask_flat = rearrange(x_mask, 'b i j -> b (i j)')
-
-                m = msa_cross_attn(
-                    m,
-                    mask = msa_mask,
-                    context = x,
-                    context_mask = x_mask_flat
-                ) + m
-
-                x = cross_attn(
-                    x,
-                    mask = x_mask_flat,
-                    context = m,
-                    context_mask = msa_mask
-                ) + x
-
-                x = rearrange(x, 'b (i j) d -> b i j d', i = n)
-
-            # feedforwards
-
-            x = ff(x) + x
-
-            if exists(msa):
-                m = msa_ff(m) + m
-
-        x = self.norm(x)
+        x, m = self.net(x, m, mask = x_mask, msa_mask = msa_mask)
 
+        x = rearrange(x, 'b (h w) d -> b h w d', h = n)
         x = (x + rearrange(x, 'b i j d -> b j i d')) * 0.5  # symmetrize
         return self.to_distogram_logits(x)
